import pika 
import cv2
import time

class AMQ_BrokerConfig:
    host = 'voicevon.vicp.io'
    port = 5672  
    virtual_host = '/'
    uid = 'von'
    password = 'von1970'


class Amq_Message():

    def __init__(self, method, body) -> None:
        self.method = method
        self.body = body

class Amq_Subscriber():
    def __init__(self, queue_name: str, channel ) -> None:
        self.queue_name = queue_name
        self.unacked_messages = [Amq_Message(None,None)]
        self.unacked_messages.clear()

        self.channel = channel
        var_callback = self.CopyToUnackedMessage
        self.channel.basic_consume(queue=queue_name, on_message_callback=var_callback, auto_ack=False)

    def CopyToUnackedMessage(self, ch, method, properties, body):
        unacked = Amq_Message(method, body)
        self.unacked_messages.append(unacked)

    def FetchMessage(self) -> str:
        if len(self.unacked_messages) > 0:
            result = self.unacked_messages[0].body
            tag = self.unacked_messages[0].method.delivery_tag
            self.channel.basic_ack(tag)
            self.unacked_messages.pop(0)
        else:
            result = None
        return result

class AmqAgent():
    '''
    To learn:  What is channel indeed ?
    '''
    def __init__(self) -> None:
        self.subscribers=[]

    def connect_to_broker(self, broker_config: AMQ_BrokerConfig) -> None:
        self.serverConfig = broker_config
        self.reconnect_to_broker()
        self.prefeched_message = None
        self.delivery_tag = None

    def process_data_events(self, time_limit=0):
        if self.channel.connection.is_closed:
            print("AmqAgent::process_data_events()  Connection is closed. Trying to reconnect ")
            input("press any key to reconnect")
            self.reconnect_to_broker()
        self.blocking_connection.process_data_events(time_limit)

    def reconnect_to_broker(self):
        broker_config = self.serverConfig
        credentials = pika.PlainCredentials(broker_config.uid, broker_config.password)
        parameters = pika.ConnectionParameters(host=broker_config.host,
                                    port=broker_config.port,
                                    virtual_host=broker_config.virtual_host,
                                    credentials=credentials)
        try:
            print("Start to connect to RabbitMQ broker.")
            self.blocking_connection = pika.BlockingConnection(parameters)
            self.channel = self.blocking_connection.channel()
            self.channel.basic_qos(prefetch_count=10)
            print("Connected RabbitMQ broker!")
        except Exception as e:
            print(e)

    def Publish(self, exchange_name: str, queue_name: str, payload: str):
        # if not (queue_name in self.declaed_queues):
        #     self.channel.queue_declare(queue=queue_name)
        #     self.declaed_queues.append(queue_name)

        self.channel.basic_publish(exchange = exchange_name,
                        routing_key = queue_name,
                        body = payload)

    def PublishBatch(self, queue_name:str, payloads:list):
        print("[Info] RabbitMqAgent.PublishBatch(), queue_name=, payload= ", queue_name, payloads)
        # if not (queue_name in self.declaed_queues):
        #     self.channel.queue_declare(queue=queue_name)
        #     self.declaed_queues.append(queue_name)
        for pp in payloads:
            self.channel.basic_publish(exchange = '',
                            routing_key = queue_name,
                            body = pp)

    def publish_cv_image(self, queue_name:str, cv_image):
        # Convert received message into Numpy array
        # jpg = np.frombuffer(RECEIVEDMESSAGE, dtype=np.uint8)

        # # JPEG-decode back into original frame - which is actually a Numpy array
        # im = cv2.imdecode(jpg, cv2.IMREAD_UNCHANGED)
        if not (queue_name in self.declaed_queues):
            self.channel.queue_declare(queue=queue_name)
            self.declaed_queues.append(queue_name)

        is_success, img_encode = cv2.imencode(".jpg", cv_image)
        if is_success:
            img_pub = img_encode.tobytes()
            # self.client.publish(topic, img_pub, retain=retain)
            self.channel.basic_publish(exchange = '',
                        routing_key = queue_name,
                        body = img_pub)

    def __FindSubscriber(self, queue_name:str) -> Amq_Subscriber:
        for subscriber in self.subscribers:
            if subscriber.queue_name == queue_name:
                return subscriber
        return None

    def fetch_message_payload(self, queue_name:str) -> str:
        subscriber = self.__FindSubscriber(queue_name)
        if subscriber is None:
            print("no subscriber is found")
            return None
        else:
            return subscriber.FetchMessage()

    def Subscribe(self, queue_name:str):
        '''
        use fetch_message_payload() to check out message.body
        '''
        new_subscriber = Amq_Subscriber(queue_name, self.channel)
        self.subscribers.append(new_subscriber)

        # new_subscriber.channel = self.channel
        # var_callback = callback
        # if callback is None:
        #     var_callback = new_subscriber.CopyToUnackedMessage
        # new_subscriber.channel.basic_consume(queue=queue_name, on_message_callback=var_callback, auto_ack=False)

    def RabbitMQ_publish_tester(self):
        i = 0
        while True:
            self.channel.basic_publish(exchange='',
                            routing_key='gobot_x2134_house',
                            body = str(i))
            self.channel.basic_publish(exchange='',
                            routing_key='gobot_x2134_arm',
                            body = str(i))
            print(" [x] Sent ",i)
            i += 1  
            time.sleep(2)

g_amq = AmqAgent()
g_amq_broker_config = AMQ_BrokerConfig()


if __name__ == '__main__':
    g_amq.connect_to_broker(g_amq_broker_config)

    # img = cv2.imread("nocommand.jpg")
    # g_amq.publish_cv_image("test" , img)
    # g_amq_broker_config.host = 'localhost'

    g_amq.Subscribe(queue_name='twh_deposit')
    g_amq.Subscribe(queue_name='twh_withdraw')
    count = 0
    g_amq.process_data_events()
    while True:
        g_amq.process_data_events()
        # g_amq.process_data_events()
        time.sleep(0.9)
        xx = g_amq.fetch_message_payload('twh_deposit')
        # xx = ss.FetchMessage()
        if xx is not None:
            print(xx)
            time.sleep(1)

        xx = g_amq.fetch_message_payload('twh_withdraw')
        if xx is not None:
            # g_amq.process_data_events()
            if count>1:
                print("withdraw is empty................................", count)
            count =0
            # print('final fetchd.', xx)
        else:
            # print("fetched   None", count)
            count += 1
        # time.sleep(1)


    


class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Person(metaclass=Singleton):
    def __init__(self, name):
        self.name = name

    def rename(self, name):
        self.name = name


if __name__ == "__main__":
    me = Person('me')
    print (me.name)

    you = Person('you')
    print(you.name)
    print(me.name)
    
    you.rename('who')
    print(you.name)
    print(me.name)







# sudo chmod 666 /dev/ttyUSB0
# roslaunch faze4_moveit demo.launch rviz_tutorial:=true


# import rospy 
# from std_msgs.msg import String
# from sensor_msgs.msg import JointState

from abc import abstractmethod
import serial, time
import serial.tools.list_ports

import sys
sys.path.append('/home/znkzjs/bot/python')
# from singleton import Singleton

from enum import Enum
#TODO: add mqtt
# from mqtt_helper import  g_mqtt

class HARD_ROBOT_ONLINE_LEVEL(Enum):
    OFF_LINE = 1
    ONLINE_AS_REPRAP = 2
    ONLINE_AS_SOWER = 3
    HOMED = 4

# class ReprapArm(metaclass=Singleton):
class ReprapArm():
    '''
    Can communacate with reprap_arm,  for example:  Marlin
    Send gcode to serial, confirm Reprap got the gcode correctly.
    '''
    def __init__(self):
        '''
        This will create an object, But will not connect to Marlin
        '''
        self.is_connected = False
        self.__serialport = serial.Serial()
        self.__counter = 0
        self.__echo_is_on = False 

        self.state = HARD_ROBOT_ONLINE_LEVEL.OFF_LINE

    def try_to_find_ports(self, portname, do_list_ports = False):
        myports = [tuple(p) for p in list(serial.tools.list_ports.comports())]
        found_com_port = False

        for a,b,c in myports:
            if do_list_ports:
                print(a)
                print(b)
                print(c)
            if a == portname:
                found_com_port = True
        return found_com_port

    def connect_reprap_controller(self, portname, baudrate):
        self.__serialport.port = portname
        self.__serialport.baudrate  = baudrate
        self.__serialport.timeout = 1
        self.__serialport.writeTimeout = 2

        if self.try_to_find_ports(portname):
            self.__serialport.open()
            if self.__echo_is_on:
                print ('Reprap_host::Serial port is opened.')
            while True:
                xx = self.__serialport.readline()
                mm = bytes.decode(xx)
                if (mm == ''):
                    self.is_connected = True
                    return True
        else:
            # Can not find 'dev/ttyUSB0'
            print('Can not find %s ' %portname)
            return False
        self.state = HARD_ROBOT_ONLINE_LEVEL.ONLINE_AS_REPRAP

    def set_echo_on(self, is_on):
        self.__echo_is_on = is_on

    def __send_gcode_mcode(self, raw_gcode):
        '''
        raw_gcode is a pure string, not include formated 
        '''
        self.__counter += 1
        self.__serialport.write(str.encode(raw_gcode +'\r\n'))
        if self.__echo_is_on:
            print ('>>> %s' % raw_gcode)
        got_ok = False
        while not got_ok:
            response_a = self.__serialport.readline()
            response = bytes.decode(response_a)
            if(response == 'ok\n'):
                got_ok = True
                if self.__echo_is_on:
                    print ('OK')
            elif (response ==''):
                time.sleep(0.1)
            elif self.__echo_is_on:
                print("<<< " + response)
    @abstractmethod
    def set_joints_angle_in_degree(self, IK_dict):
        print('set_joints_angle_in_degree not implatemented')

    def set_fan_speed(self, speed):
        '''
        speed from 0 to 255
        '''
        self.__send_gcode_mcode('M106 S' + str(speed))

    def disable_motor_sleep(self):
        self.__send_gcode_mcode('M84 S0')  #Disable sleep

    def print_homing_sensor_states(self):
        '''
        print out status of all joints home sensor.
        include joint5? I forgot it, depend on Marlin firmware. 
        '''
        self.__send_gcode_mcode('M84')
        while True:
            self.set_echo_on(True)
            self.__send_gcode_mcode('M119')
            # rospy.sleep(1)
            time.sleep(1)

    def allow_cold_extrusion(self):
        self.__send_gcode_mcode('M302 S0') # Allow extrusion at any temperature
    
    def wait_for_movement_finsished(self):
        self.__send_gcode_mcode('M400')

    def home(self, home_x=False, home_y=False, home_z=False):
        if home_x:
            self.__send_gcode_mcode('G28 X')
        if home_y:
            self.__send_gcode_mcode('G28 Y')
        if home_z:
            self.__send_gcode_mcode('G28 Z')

    def move_to_xyz(self, x, y, z=None, speed_mm_per_min=None):
        gcode = 'G1'
        if x is not None:
            gcode += 'X' + str(x)
        if y is not None:
            gcode += 'Y' + str(y)
        if z is not None:
            gcode += 'Z' + str(z)
        if speed_mm_per_min is not None:
            gcode += 'F' + str(speed_mm_per_min)
        self.__send_gcode_mcode(gcode)

    def bridge_send_gcode_mcode(self, the_code):
        self.__send_gcode_mcode(the_code)
    
    def set_servo_position(self, pwm_id=0, angle=11):
        mcode = 'M280 P' + str(pwm_id) +  ' S' + str(angle)
        print(mcode)
        self.__send_gcode_mcode(mcode)


    @abstractmethod
    def home_all_joints(self):
        print('[Warning] This is an abstract method')
        
    
class TestArm(ReprapArm):
    pass

if __name__ == "__main__":
    test = TestArm()
    test.set_echo_on(True)
    test.connect_reprap_controller('/dev/ttyUSB0', 115200)
    test.bridge_send_gcode_mcode('M114')
    test.home(home_y=True)

